from that_depends import providers
from that_depends.container import BaseContainer
from that_depends.injection import Provide, inject
from that_depends.providers import container_context
from that_depends.providers.context_resources import fetch_context_item


__all__ = [
    "container_context",
    "fetch_context_item",
    "providers",
    "BaseContainer",
    "inject",
    "Provide",
]



import typing

from litestar import Controller, Litestar, Router, get
from litestar.di import Provide
from litestar.status_codes import HTTP_200_OK
from litestar.testing import TestClient

from that_depends import BaseContainer, providers


def bool_fn(value: bool) -> bool:
    return value


def str_fn() -> str:
    return ""


def list_fn() -> list[str]:
    return ["some"]


def int_fn() -> int:
    return 1


class DIContainer(BaseContainer):
    bool_fn = providers.Factory(bool_fn, value=False)
    str_fn = providers.Factory(str_fn)
    list_fn = providers.Factory(list_fn)
    int_fn = providers.Factory(int_fn)


class MyController(Controller):
    path = "/controller"
    dependencies = {"controller_dependency": Provide(DIContainer.list_fn)}  # noqa: RUF012

    @get(path="/handler", dependencies={"local_dependency": Provide(DIContainer.int_fn)})
    async def my_route_handler(
        self,
        app_dependency: bool,
        router_dependency: str,
        controller_dependency: list[str],
        local_dependency: int,
    ) -> dict[str, typing.Any]:
        return {
            "app_dependency": app_dependency,
            "router_dependency": router_dependency,
            "controller_dependency": controller_dependency,
            "local_dependency": local_dependency,
        }


my_router = Router(
    path="/router",
    dependencies={"router_dependency": Provide(DIContainer.str_fn)},
    route_handlers=[MyController],
)

# on the app
app = Litestar(route_handlers=[my_router], dependencies={"app_dependency": Provide(DIContainer.bool_fn)}, debug=True)


def test_litestar_di() -> None:
    with TestClient(app=app) as client:
        response = client.get("/router/controller/handler")
        assert response.status_code == HTTP_200_OK, response.text
        assert response.json() == {
            "app_dependency": False,
            "controller_dependency": ["some"],
            "local_dependency": 1,
            "router_dependency": "",
        }



import random
from dataclasses import dataclass, field

import pytest

from that_depends import providers
from that_depends.providers.attr_getter import _get_value_from_object_by_dotted_path


@dataclass
class Nested2:
    some_const = 144


@dataclass
class Nested1:
    nested2_attr: Nested2 = field(default_factory=Nested2)


@dataclass
class Settings:
    some_str_value: str = "some_string_value"
    some_int_value: int = 3453621
    nested1_attr: Nested1 = field(default_factory=Nested1)


@dataclass
class NestingTestDTO: ...


@pytest.fixture
def some_settings_provider() -> providers.Singleton[Settings]:
    return providers.Singleton(Settings)


def test_attr_getter_with_zero_attribute_depth(some_settings_provider: providers.Singleton[Settings]) -> None:
    attr_getter = some_settings_provider.some_str_value
    assert attr_getter.sync_resolve() == Settings().some_str_value


def test_attr_getter_with_more_than_zero_attribute_depth(some_settings_provider: providers.Singleton[Settings]) -> None:
    attr_getter = some_settings_provider.nested1_attr.nested2_attr.some_const
    assert attr_getter.sync_resolve() == Nested2().some_const


@pytest.mark.parametrize(
    ("field_count", "test_field_name", "test_value"),
    [(1, "test_field", "sdf6fF^SF(FF*4ffsf"), (5, "nested_field", -252625), (50, "50_lvl_field", 909234235)],
)
def test_nesting_levels(field_count: int, test_field_name: str, test_value: str | int) -> None:
    obj = NestingTestDTO()
    fields = [f"field_{i}" for i in range(1, field_count + 1)]
    random.shuffle(fields)

    attr_path = ".".join(fields) + f".{test_field_name}"
    obj_copy = obj

    while fields:
        field_name = fields.pop(0)
        setattr(obj_copy, field_name, NestingTestDTO())
        obj_copy = obj_copy.__getattribute__(field_name)

    setattr(obj_copy, test_field_name, test_value)

    attr_value = _get_value_from_object_by_dotted_path(obj, attr_path)
    assert attr_value == test_value


def test_attr_getter_with_invalid_attribute(some_settings_provider: providers.Singleton[Settings]) -> None:
    with pytest.raises(AttributeError):
        some_settings_provider.nested1_attr.nested2_attr.__some_private__  # noqa: B018
    with pytest.raises(AttributeError):
        some_settings_provider.nested1_attr.__another_private__  # noqa: B018
    with pytest.raises(AttributeError):
        some_settings_provider.nested1_attr._final_private_  # noqa: B018



import datetime
import logging
import typing

import pytest

from tests.container import create_async_resource, create_sync_resource
from that_depends import BaseContainer, providers


logger = logging.getLogger(__name__)
global_state_for_selector: typing.Literal["sync_resource", "async_resource", "missing"] = "sync_resource"


class SelectorState:
    def __init__(self) -> None:
        self.selector_state: typing.Literal["sync_resource", "async_resource", "missing"] = "sync_resource"

    def get_selector_state(self) -> typing.Literal["sync_resource", "async_resource", "missing"]:
        return self.selector_state


selector_state = SelectorState()


class DIContainer(BaseContainer):
    sync_resource = providers.Resource(create_sync_resource)
    async_resource = providers.Resource(create_async_resource)
    selector: providers.Selector[datetime.datetime] = providers.Selector(
        selector_state.get_selector_state,
        sync_resource=sync_resource,
        async_resource=async_resource,
    )


async def test_selector_provider_async() -> None:
    selector_state.selector_state = "async_resource"
    selected = await DIContainer.selector()
    async_resource = await DIContainer.async_resource()

    assert selected == async_resource


async def test_selector_provider_async_missing() -> None:
    selector_state.selector_state = "missing"
    with pytest.raises(RuntimeError, match="No provider matches"):
        await DIContainer.selector()


async def test_selector_provider_sync() -> None:
    selector_state.selector_state = "sync_resource"
    selected = DIContainer.selector.sync_resolve()
    sync_resource = DIContainer.sync_resource.sync_resolve()

    assert selected == sync_resource


async def test_selector_provider_sync_missing() -> None:
    selector_state.selector_state = "missing"
    with pytest.raises(RuntimeError, match="No provider matches"):
        DIContainer.selector.sync_resolve()


async def test_selector_provider_overriding() -> None:
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    DIContainer.selector.override(now)
    selected_async = await DIContainer.selector()
    selected_sync = DIContainer.selector.sync_resolve()
    assert selected_async == selected_sync == now

    DIContainer.reset_override()
    selector_state.selector_state = "sync_resource"
    selected = DIContainer.selector.sync_resolve()
    sync_resource = DIContainer.sync_resource.sync_resolve()
    assert selected == sync_resource



import typing

from that_depends.providers.base import AbstractFactory, AbstractProvider


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class Factory(AbstractFactory[T_co]):
    __slots__ = "_factory", "_args", "_kwargs", "_override"

    def __init__(self, factory: type[T_co] | typing.Callable[P, T_co], *args: P.args, **kwargs: P.kwargs) -> None:
        super().__init__()
        self._factory: typing.Final = factory
        self._args: typing.Final = args
        self._kwargs: typing.Final = kwargs
        self._override = None

    async def async_resolve(self) -> T_co:
        if self._override:
            return typing.cast(T_co, self._override)

        return self._factory(
            *[await x.async_resolve() if isinstance(x, AbstractProvider) else x for x in self._args],
            **{k: await v.async_resolve() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()},
        )

    def sync_resolve(self) -> T_co:
        if self._override:
            return typing.cast(T_co, self._override)

        return self._factory(
            *[x.sync_resolve() if isinstance(x, AbstractProvider) else x for x in self._args],
            **{k: v.sync_resolve() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()},
        )


class AsyncFactory(AbstractFactory[T_co]):
    __slots__ = "_factory", "_args", "_kwargs", "_override"

    def __init__(self, factory: typing.Callable[P, typing.Awaitable[T_co]], *args: P.args, **kwargs: P.kwargs) -> None:
        self._factory: typing.Final = factory
        self._args: typing.Final = args
        self._kwargs: typing.Final = kwargs
        self._override = None

    async def async_resolve(self) -> T_co:
        if self._override:
            return typing.cast(T_co, self._override)

        return await self._factory(
            *[await x.async_resolve() if isinstance(x, AbstractProvider) else x for x in self._args],
            **{k: await v.async_resolve() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()},
        )

    def sync_resolve(self) -> typing.NoReturn:
        msg = "AsyncFactory cannot be resolved synchronously"
        raise RuntimeError(msg)



import typing

from that_depends.providers.base import AbstractProvider


T_co = typing.TypeVar("T_co", covariant=True)


class List(AbstractProvider[list[T_co]]):
    __slots__ = ("_providers",)

    def __init__(self, *providers: AbstractProvider[T_co]) -> None:
        super().__init__()
        self._providers: typing.Final = providers

    async def async_resolve(self) -> list[T_co]:
        return [await x.async_resolve() for x in self._providers]

    def sync_resolve(self) -> list[T_co]:
        return [x.sync_resolve() for x in self._providers]

    async def __call__(self) -> list[T_co]:
        return await self.async_resolve()


class Dict(AbstractProvider[dict[str, T_co]]):
    __slots__ = ("_providers",)

    def __init__(self, **providers: AbstractProvider[T_co]) -> None:
        super().__init__()
        self._providers: typing.Final = providers

    async def async_resolve(self) -> dict[str, T_co]:
        return {key: await provider.async_resolve() for key, provider in self._providers.items()}

    def sync_resolve(self) -> dict[str, T_co]:
        return {key: provider.sync_resolve() for key, provider in self._providers.items()}



import dataclasses

import pydantic

from that_depends import BaseContainer, providers


@dataclasses.dataclass(kw_only=True, slots=True)
class SingletonFactory:
    dep1: str


class Settings(pydantic.BaseModel):
    some_setting: str = "some_value"
    other_setting: str = "other_value"


class DIContainer(BaseContainer):
    settings: Settings = providers.Singleton(Settings).cast
    singleton = providers.Singleton(SingletonFactory, dep1=settings.some_setting)


async def test_singleton_provider() -> None:
    singleton1 = await DIContainer.singleton()
    singleton2 = await DIContainer.singleton()
    singleton3 = DIContainer.singleton.sync_resolve()
    await DIContainer.singleton.tear_down()
    singleton4 = DIContainer.singleton.sync_resolve()

    assert singleton1 is singleton2 is singleton3
    assert singleton4 is not singleton1

    await DIContainer.tear_down()


async def test_singleton_attr_getter() -> None:
    singleton1 = await DIContainer.singleton()

    assert singleton1.dep1 == Settings().some_setting

    await DIContainer.tear_down()



import datetime
import typing

import fastapi
from starlette import status
from starlette.testclient import TestClient

from tests import container
from that_depends.providers import DIContextMiddleware


app = fastapi.FastAPI()
app.add_middleware(DIContextMiddleware)


@app.get("/")
async def read_root(
    dependency: typing.Annotated[
        container.DependentFactory,
        fastapi.Depends(container.DIContainer.dependent_factory),
    ],
    free_dependency: typing.Annotated[
        container.FreeFactory,
        fastapi.Depends(container.DIContainer.resolver(container.FreeFactory)),
    ],
    singleton: typing.Annotated[
        container.SingletonFactory,
        fastapi.Depends(container.DIContainer.singleton),
    ],
    singleton_attribute: typing.Annotated[bool, fastapi.Depends(container.DIContainer.singleton.dep1)],
) -> datetime.datetime:
    assert dependency.sync_resource == free_dependency.dependent_factory.sync_resource
    assert dependency.async_resource == free_dependency.dependent_factory.async_resource
    assert singleton.dep1 is True
    assert singleton_attribute is True
    return dependency.async_resource


client = TestClient(app)


async def test_read_main() -> None:
    response = client.get("/")
    assert response.status_code == status.HTTP_200_OK
    assert (
        datetime.datetime.fromisoformat(response.json().replace("Z", "+00:00"))
        == await container.DIContainer.async_resource()
    )



import typing

from that_depends.providers.base import AbstractProvider


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class Object(AbstractProvider[T_co]):
    __slots__ = ("_obj",)

    def __init__(self, obj: T_co) -> None:
        super().__init__()
        self._obj: typing.Final = obj

    async def async_resolve(self) -> T_co:
        return self._obj

    def sync_resolve(self) -> T_co:
        return self._obj



import datetime
import logging
import typing
import uuid
from contextlib import AsyncExitStack

import pytest

from that_depends import BaseContainer, Provide, fetch_context_item, inject, providers
from that_depends.providers import container_context
from that_depends.providers.base import ResourceContext


logger = logging.getLogger(__name__)


def create_sync_context_resource() -> typing.Iterator[str]:
    logger.info("Resource initiated")
    yield f"sync {uuid.uuid4()}"
    logger.info("Resource destructed")


async def create_async_context_resource() -> typing.AsyncIterator[str]:
    logger.info("Async resource initiated")
    yield f"async {uuid.uuid4()}"
    logger.info("Async resource destructed")


class DIContainer(BaseContainer):
    sync_context_resource = providers.ContextResource(create_sync_context_resource)
    async_context_resource = providers.ContextResource(create_async_context_resource)
    dynamic_context_resource = providers.Selector(
        lambda: fetch_context_item("resource_type") or "sync",
        sync=sync_context_resource,
        async_=async_context_resource,
    )


@pytest.fixture(autouse=True)
async def _clear_di_container() -> typing.AsyncIterator[None]:
    try:
        yield
    finally:
        await DIContainer.tear_down()


@pytest.fixture(params=[DIContainer.sync_context_resource, DIContainer.async_context_resource])
def context_resource(request: pytest.FixtureRequest) -> providers.ContextResource[str]:
    return typing.cast(providers.ContextResource[str], request.param)


@pytest.fixture
def sync_context_resource() -> providers.ContextResource[str]:
    return DIContainer.sync_context_resource


@pytest.fixture
def async_context_resource() -> providers.ContextResource[str]:
    return DIContainer.async_context_resource


async def test_context_resource_without_context_init(
    context_resource: providers.ContextResource[str],
) -> None:
    with pytest.raises(RuntimeError, match="Context is not set. Use container_context"):
        await context_resource.async_resolve()

    with pytest.raises(RuntimeError, match="Context is not set. Use container_context"):
        context_resource.sync_resolve()


@container_context()
async def test_context_resource(context_resource: providers.ContextResource[str]) -> None:
    context_resource_result = await context_resource()

    assert await context_resource() is context_resource_result


@container_context()
def test_sync_context_resource(sync_context_resource: providers.ContextResource[str]) -> None:
    context_resource_result = sync_context_resource.sync_resolve()

    assert sync_context_resource.sync_resolve() is context_resource_result


async def test_async_context_resource_in_sync_context(async_context_resource: providers.ContextResource[str]) -> None:
    with pytest.raises(RuntimeError, match="AsyncResource cannot be resolved in an sync context."), container_context():
        await async_context_resource()


async def test_context_resource_different_context(
    context_resource: providers.ContextResource[datetime.datetime],
) -> None:
    async with container_context():
        context_resource_instance1 = await context_resource()

    async with container_context():
        context_resource_instance2 = await context_resource()

    assert context_resource_instance1 is not context_resource_instance2


async def test_context_resource_included_context(
    context_resource: providers.ContextResource[datetime.datetime],
) -> None:
    async with container_context():
        context_resource_instance1 = await context_resource()
        async with container_context():
            context_resource_instance2 = await context_resource()

        context_resource_instance3 = await context_resource()

    assert context_resource_instance1 is not context_resource_instance2
    assert context_resource_instance1 is context_resource_instance3


async def test_context_resources_overriding(context_resource: providers.ContextResource[str]) -> None:
    context_resource_mock = datetime.datetime.now(tz=datetime.timezone.utc)
    context_resource.override(context_resource_mock)

    context_resource_result = await context_resource()
    context_resource_result2 = context_resource.sync_resolve()
    assert context_resource_result is context_resource_result2 is context_resource_mock

    DIContainer.reset_override()
    with pytest.raises(RuntimeError, match="Context is not set. Use container_context"):
        await context_resource()


async def test_context_resources_init_and_tear_down() -> None:
    await DIContainer.init_resources()
    await DIContainer.tear_down()


def test_context_resources_wrong_providers_init() -> None:
    with pytest.raises(RuntimeError, match="ContextResource must be generator function"):
        providers.ContextResource(lambda: None)  # type: ignore[arg-type,return-value]


async def test_context_resource_with_dynamic_resource() -> None:
    async with container_context({"resource_type": "sync"}):
        assert (await DIContainer.dynamic_context_resource()).startswith("sync")

    async with container_context({"resource_type": "async_"}):
        assert (await DIContainer.dynamic_context_resource()).startswith("async")

    async with container_context():
        assert (await DIContainer.dynamic_context_resource()).startswith("sync")


async def test_early_exit_of_container_context() -> None:
    with pytest.raises(RuntimeError, match="Context is not set, call ``__aenter__`` first"):
        await container_context().__aexit__(None, None, None)
    with pytest.raises(RuntimeError, match="Context is not set, call ``__enter__`` first"):
        container_context().__exit__(None, None, None)


async def test_resource_context_early_teardown() -> None:
    context: ResourceContext[str] = ResourceContext(is_async=True)
    assert context.context_stack is None
    context.sync_tear_down()
    assert context.context_stack is None


async def test_teardown_sync_container_context_with_async_resource() -> None:
    resource_context: ResourceContext[typing.Any] = ResourceContext(is_async=True)
    resource_context.context_stack = AsyncExitStack()
    with pytest.raises(RuntimeError, match="Cannot tear down async context in sync mode"):
        resource_context.sync_tear_down()


async def test_sync_container_context_with_different_stack() -> None:
    @container_context()
    @inject
    def some_injected(depth: int, val: str = Provide[DIContainer.sync_context_resource]) -> str:
        if depth > 1:
            return val
        return some_injected(depth + 1)

    some_injected(1)


async def test_async_container_context_with_different_stack() -> None:
    @container_context()
    @inject
    async def some_injected(depth: int, val: str = Provide[DIContainer.async_context_resource]) -> str:
        if depth > 1:
            return val
        return await some_injected(depth + 1)

    await some_injected(1)



import logging
import typing

from that_depends import BaseContainer, providers


logger = logging.getLogger(__name__)


def create_sync_resource() -> typing.Iterator[list[typing.Any]]:
    logger.debug("Resource initiated")
    yield []
    logger.debug("Resource destructed")


async def create_async_resource() -> typing.AsyncIterator[dict[str, typing.Any]]:
    logger.debug("Async resource initiated")
    yield {}
    logger.debug("Async resource destructed")


class TrickyDIContainer(BaseContainer):
    singleton = providers.Singleton(list)
    sync_resource = providers.Resource(create_sync_resource)
    async_resource = providers.Resource(create_async_resource)


async def test_singleton_with_empty_list() -> None:
    singleton1 = await TrickyDIContainer.singleton()
    singleton2 = await TrickyDIContainer.singleton()
    assert singleton1 is singleton2


async def test_resources_with_empty_list() -> None:
    sync_resource1 = await TrickyDIContainer.sync_resource()
    sync_resource2 = await TrickyDIContainer.sync_resource()
    assert sync_resource1 is sync_resource2

    async_resource1 = await TrickyDIContainer.async_resource()
    async_resource2 = await TrickyDIContainer.async_resource()
    assert async_resource1 is async_resource2

    await TrickyDIContainer.tear_down()



import typing

from that_depends.providers.base import AbstractProvider


T_co = typing.TypeVar("T_co", covariant=True)


class Selector(AbstractProvider[T_co]):
    __slots__ = "_selector", "_providers", "_override"

    def __init__(self, selector: typing.Callable[[], str], **providers: AbstractProvider[T_co]) -> None:
        super().__init__()
        self._selector: typing.Final = selector
        self._providers: typing.Final = providers
        self._override = None

    async def async_resolve(self) -> T_co:
        if self._override:
            return typing.cast(T_co, self._override)

        selected_key: typing.Final = self._selector()
        if selected_key not in self._providers:
            msg = f"No provider matches {selected_key}"
            raise RuntimeError(msg)
        return await self._providers[selected_key].async_resolve()

    def sync_resolve(self) -> T_co:
        if self._override:
            return typing.cast(T_co, self._override)

        selected_key: typing.Final = self._selector()
        if selected_key not in self._providers:
            msg = f"No provider matches {selected_key}"
            raise RuntimeError(msg)
        return self._providers[selected_key].sync_resolve()



import logging

import pytest

from that_depends import BaseContainer, providers


logger = logging.getLogger(__name__)


class DIContainer(BaseContainer):
    pass


async def test_init_async_resources() -> None:
    await DIContainer.init_async_resources()


def test_wrong_deprecated_providers_init() -> None:
    with pytest.raises(RuntimeError, match="AsyncContextResource must be generator function"):
        providers.AsyncContextResource(lambda: None)  # type: ignore[arg-type,return-value]

    with pytest.raises(RuntimeError, match="AsyncResource must be generator function"):
        providers.AsyncResource(lambda: None)  # type: ignore[arg-type,return-value]



from that_depends import BaseContainer, providers


instance = "some string"


class DIContainer(BaseContainer):
    instance = providers.Object(instance)


async def test_object_provider() -> None:
    instance1 = await DIContainer.instance()
    instance2 = DIContainer.instance.sync_resolve()

    assert instance1 is instance2 is instance



import dataclasses

import pytest

from tests import container
from tests.container import DIContainer


@dataclasses.dataclass(kw_only=True, slots=True)
class WrongFactory:
    some_dep: int = 1
    not_existing_name: container.DependentFactory
    sync_resource: str


async def test_dependency_resolver() -> None:
    resolver = DIContainer.resolver(container.FreeFactory)
    dep1 = await resolver()
    dep2 = await DIContainer.resolve(container.FreeFactory)

    assert dep1
    assert dep2
    assert dep1 is not dep2


async def test_dependency_resolver_failed() -> None:
    resolver = DIContainer.resolver(WrongFactory)
    with pytest.raises(RuntimeError, match="Provider is not found, field_name='not_existing_name'"):
        await resolver()



import dataclasses
import typing

import pytest

from tests import container
from that_depends import BaseContainer, providers


@dataclasses.dataclass(kw_only=True, slots=True)
class InjectedFactories:
    sync_factory: typing.Callable[..., container.DependentFactory]
    async_factory: typing.Callable[..., typing.Coroutine[typing.Any, typing.Any, container.DependentFactory]]


class DIContainer(BaseContainer):
    sync_resource = providers.Resource(container.create_sync_resource)
    async_resource = providers.Resource(container.create_async_resource)

    simple_factory = providers.Factory(container.SimpleFactory, dep1="text", dep2=123)
    dependent_factory = providers.Factory(
        container.DependentFactory,
        simple_factory=simple_factory.cast,
        sync_resource=sync_resource.cast,
        async_resource=async_resource.cast,
    )
    injected_factories = providers.Factory(
        InjectedFactories,
        sync_factory=dependent_factory.sync_provider,
        async_factory=dependent_factory.provider,
    )


async def test_async_provider() -> None:
    injected_factories = await DIContainer.injected_factories()
    instance1 = await injected_factories.async_factory()
    instance2 = await injected_factories.async_factory()

    assert isinstance(instance1, container.DependentFactory)
    assert isinstance(instance2, container.DependentFactory)
    assert instance1 is not instance2

    await DIContainer.tear_down()


async def test_sync_provider() -> None:
    injected_factories = await DIContainer.injected_factories()
    with pytest.raises(RuntimeError, match="AsyncResource cannot be resolved synchronously"):
        injected_factories.sync_factory()

    await DIContainer.init_resources()
    instance1 = injected_factories.sync_factory()
    instance2 = injected_factories.sync_factory()

    assert isinstance(instance1, container.DependentFactory)
    assert isinstance(instance2, container.DependentFactory)
    assert instance1 is not instance2

    await DIContainer.tear_down()
